;; Generated package description from swiper.el
(define-package "swiper" "0.2.0" "Isearch with an overview. Oh, man!" '((emacs "24.1")) :url "https://github.com/abo-abo/swiper" :keywords '("matching"))
